
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include "basis_equation.h"
#include "boundary.h"
#include "boundary_state.h"
#include "checkpoint.h"
#include "combinat.h"
#include "common_startup.h"
#include "convexity.h"
#include "debug.h"
#include "facet_list.h"
#include "globaldef.h"
#include "gparams.h"
#include "matroid.h"
#include "nalloc.h"
#include "nodedef.h"
#include "nu_boundary.h"
#include "nu_signed_index.h"
#include "oms_opt.h"
#include "oms_types.h"
#include "output.h"
#include "parser.h"
#include "pluecker.h"
#include "profile.h"
#include "runtime_options.h"
#include "sa32_array.h"
#include "shutdown.h"
#ifdef USE_ZRAM
#include "z_checkpoint.h"
#endif
#include "int_to_symbol.h"
extern int oms_GlobalTag,oms_NodeTag;

/**********************************************************************/
/* static function prototypes */
static bool read_chirotope(na_node_p global, na_node_p local, sarray_p chirotope, startopt_t startopt);
static bool read_boundary(na_node_p global, na_node_p local, sarray_p boundary, oms_hook read_boundary_hook,startopt_t startopt);
static void set_root_basis(na_node_p global, na_node_p local,params_p params);

/* global functions resolved at link time */
extern bool SetAndConsequences(na_node_p, na_node_p, int, int);


/***********************************************************************/

/*
I'm not sure this works, since it didn't compile!
db
*/
extern na_node_p cs_restart(params_p params){
#ifdef USE_ZRAM
  na_node_p new_global;
  params_p restored_params;
  err_debug(DB_STARTUP,("init_global","restarting from checkpoint %s",
			params->restartname));
      
  new_global=(na_node_p)z_check_restart(params->restartname,
					    params->checkpointname,
					    params->checkpointtime);
  na_fixup(new_global);
  
  restored_params=na_access(new_global,SLOT_params);
      
  restored_params->restart=1;
  return new_global;
#else
  err_fatal(HERE,"restarting only supported under ZRAM");
  return 0;
#endif
}


extern na_node_p cs_create_global(int argc, char **argv, startopt_t startopt){
  na_node_p global;
  params_p params;
  global=na_new(oms_GlobalTag);
  global=na_alloc(global,SLOT_params,sizeof(struct params_struct));
  params=na_access(global,SLOT_params);

  ro_process_args(argc,argv,params,startopt);
  return global;
}

extern na_node_p cs_init_global_tables(na_node_p global,startopt_t startopt){

  be_block_p bdy_eq,pivots;
  be_block_p PE,ridge_table;

  array_index_t *reorder_table;

  int n,r;
  int num_bases,num_eqns;
  params_p params=na_access(global,SLOT_params);


  n=params->n;
  r=params->r;
  
    /* Compute how much space we need */
  num_bases=binomial(n,r);
  num_eqns=binomial(n,r+2)*binomial(r+2,4); 

    if ((startopt & SO_pluecker) || params->extra_pluecker)
      global=be_nalloc(global,SLOT_pe,num_eqns,6,num_bases);


    global=be_nalloc(global,SLOT_bdy_eq,binomial(n,r-1),n-r+1,num_bases); 

    global=be_nalloc(global,SLOT_pivots,binomial(n,r-1),
		   (n-r+1)*(r-1),binomial(n,r-1)); 

    global=be_nalloc(global,SLOT_ridge_table,binomial(n,r-1),
		     (r-1),binomial(n,r-2)); 

    if (startopt & SO_nonuniform){
      global=ce_nalloc(global,SLOT_conv_eq,n,r);
      global=me_nalloc(global,SLOT_matroid_eq,n,r);
    }
    
    global=na_alloc(global,SLOT_facet_id,binomial(n,r-1)*sizeof(array_index_t));

    if (startopt & SO_backtrack)
      global=na_alloc(global,SLOT_reorder_table,num_bases*sizeof(array_index_t));
    
    na_fixup(global); 

    params=na_access(global,SLOT_params);
    
    {
      int i;

      array_index_t *facet_id=na_access(global,SLOT_facet_id);
      
      for (i=0; i<binomial(n,r-1); i++){
	facet_id[i]=-1;
      }

    }

    if ((startopt&SO_pluecker)  || params->extra_pluecker){
      PE=na_access(global,SLOT_pe);
      pl_init(PE, n,r);
    }

    bdy_eq=na_access(global,SLOT_bdy_eq);
    if (startopt & SO_nonuniform){
      nb_boundary_init(bdy_eq, n, r, parser.facets, params->source_facet,
		       params->sink_facet,params->diameter);
    } else {
      be_boundary_init(bdy_eq,n,r,params->source_facet,
		      params->sink_facet,params->diameter);
    } 
    
    pivots=na_access(global,SLOT_pivots);
    be_neighbour_init(pivots,n,r);

    ridge_table=na_access(global,SLOT_ridge_table);
    be_ridge_table_init(ridge_table,n,r);

    

    if (startopt & SO_nonuniform){
      be_block_p conv_eq=na_access(global,SLOT_conv_eq);
      ce_init(conv_eq,n,r);

    }

    /* this creates a spurious link time dependance for hls */
    if (startopt & SO_backtrack){
      reorder_table=na_access(global,SLOT_reorder_table);
      init_reorder_table(params->variable_order,n,r,reorder_table);
    }

    return global;
}
extern void cs_checkpoint_init(na_node_p global,params_p params){
#ifdef USE_ZRAM    

    if (params->checkpointname[0]==0 && (params->checkpointtime > 0)){
      chk_set_default_name(params->checkpointname,PATH_MAX);
    }

    if (params->checkpointtime > 0 )
      z_check_setup((z_NodeP)global,
		    params->checkpointname,
		    params->checkpointtime);
#endif    
}
extern na_node_p cs_init_global(int argc, char **argv, startopt_t startopt){
  params_p params;
  na_node_p global;

  global=cs_create_global(argc,argv,startopt);
  params=na_access(global,SLOT_params);

  if (params->restart){
    global=cs_restart(params);
  } else {
    parse_input(params);
    
    global=cs_init_global_tables(global,startopt);
    cs_checkpoint_init(global,params);
  }
  return global;
}

extern na_node_p cs_alloc_local(na_node_p global,startopt_t startopt){
  na_node_p local;
  params_p params=na_access(global,SLOT_params);

  int n=params->n;
  int r=params->r;
  
  int num_bases=binomial(n,r);

  local=na_new(oms_NodeTag);
  local=na_alloc(local,SLOT_lparams,sizeof(lparams_t));
  local=sa_nalloc(local,SLOT_chirotope,  num_bases,  0 );

  if (BITS_BASE_T < n-(r-1)){
    err_fatal(HERE,"not enough bits in base_t for bitmask %d wide",
	      n-(r-1));
  }
    
  /* used by e.g. be_check_conv */
  local=sa_nalloc(local,SLOT_pos_mask, 
		  binomial(n,r-1),n);
  local=sa_nalloc(local,SLOT_neg_mask,
		  binomial(n,r-1),n);
  
  if (startopt & SO_nonuniform) {
    local=sa_nalloc(local,SLOT_zero_mask,
		  binomial(n,r-1),n);

    local=sa_nalloc(local,SLOT_conv_pos_mask,
		  n*binomial(n,r),n);

    local=sa_nalloc(local,SLOT_conv_neg_mask,
		  n*binomial(n,r),n);

    local=sa_nalloc(local,SLOT_conv_zero_mask,
		  n*binomial(n,r),n);

    local=sa_nalloc(local,SLOT_matroid_zero_mask,binomial(n,r)*r,n);
    local=sa_nalloc(local,SLOT_matroid_nonzero_mask,binomial(n,r)*r,n);
    local=sa_nalloc(local,SLOT_facet_orientation,binomial(n,r-1),n);
  }

  local=sa_nalloc(local,SLOT_boundary,
		  binomial(n,r-1),0); 
    
  /* this only makes sense in the uniform case */
  local=rg_nalloc(local,SLOT_ridge_graph,n,r,
		  params->source_facet,
		  params->sink_facet,
		  params->diameter);  
  
  na_fixup(local);

  return local;
}

extern na_node_p cs_init_local(na_node_p global,
			       na_node_p local,
			       startopt_t startopt,
			       oms_hook bdy_hook){

  int i;
  uint32_t unknown_val;
  sarray_p pos_mask,neg_mask,zero_mask,chirotope,boundary;
  na_node_p new_local;

  params_p params=na_access(global,SLOT_params);

  chirotope=na_access(local,SLOT_chirotope);
  
  if (startopt & SO_nonuniform){
    if (params->uniform)
      unknown_val=SIGN_NONZERO;
    else 
      unknown_val=SIGN_ANY;
  }
  else
    unknown_val=0;

  for (i=0; i < sa_size(chirotope); i++)
    sa_init(chirotope,i,unknown_val);

  pos_mask=na_access(local,SLOT_pos_mask);  
  neg_mask=na_access(local,SLOT_neg_mask);  

  if (startopt & SO_nonuniform){
    zero_mask=na_access(local,SLOT_zero_mask);  
  }

  for(i=0; i<sa_size(pos_mask); i++){
    sa_init(pos_mask,i,BDY_MAYBE);
    sa_init(neg_mask,i,BDY_MAYBE);
    if (startopt & SO_nonuniform)
      sa_init(zero_mask,i,BDY_MAYBE);
  }

  if (startopt & SO_nonuniform){
    sarray_p conv_pos_mask=na_access(local,SLOT_conv_pos_mask);
    sarray_p conv_neg_mask=na_access(local,SLOT_conv_neg_mask);
    sarray_p conv_zero_mask=na_access(local,SLOT_conv_zero_mask);
    sarray_p facet_orientation=na_access(local,SLOT_facet_orientation);
    
    for(i=0; i<sa_size(conv_pos_mask); i++){
      sa_init(conv_pos_mask,i,BDY_MAYBE);
      sa_init(conv_neg_mask,i,BDY_MAYBE);
      sa_init(conv_zero_mask,i,BDY_MAYBE);
    }
    for(i=0; i<sa_size(facet_orientation); i++){
      sa_init(facet_orientation,i,SIGN_ANY);
    }

  }
  
  boundary=na_access(local,SLOT_boundary);
  for (i=0; i < sa_size(boundary); i++){
    sa_init(boundary,i,BDY_MAYBE);
  }
  
  
  if (startopt & SO_nonuniform){
    me_setup(global,local);
    set_root_basis(global,local,params);
  }

  /* Grab chirotope from parser, and verify it is consistent*/
  if (!read_chirotope(global,local,chirotope,startopt))
    return NULL;

  if (!read_boundary(global,local,boundary,bdy_hook,startopt))
    return NULL;

  params->success_count=0;
  if (params->log_input!='n'){
    printf("# instance %d\n",parser.count);
    if (params->log_input!='c'){
      output_chi_bdy(params,params->log_input,chirotope,boundary);
    }
  }
      
  new_local=na_compact(local);
  na_destroy(local);

  return new_local;
  return local;
}

static void set_root_basis(na_node_p global,
			   na_node_p local,
			   params_p params){
  
  be_block_p matroid_eq;
  sarray_p matroid_zero_mask,chirotope;
  lparams_p lparams=na_access(local,SLOT_lparams);

  array_index_t beta0=-1;
  int i;

  lparams->root_basis=-1;
  
  for (i=0; i <nss_count(parser.basis_signs)  && beta0<0; i++){
    int b;
    nu_sign_t sigma;
    bool ok;
    nu_signed_index_t si=nss_next(parser.basis_signs);
    
    b=nsi_index(si);
    sigma=nsi_sign(si);
    err_debug(DB_STARTUP,(HERE,"root_basis ? %d %c",b,sign2char(sigma)));
    if (sign_subset(sigma,SIGN_NONZERO))
      beta0=b;
  }

  if (beta0 ==-1){
    err_warning("startup",
		"no nonzero chirotope sign set in input.");
    params->first_root_basis=0;
    params->last_root_basis=binomial(params->n,params->r)-1;
  } else {
    params->first_root_basis=beta0;
    params->last_root_basis=beta0;
    lparams->root_basis=beta0;
    matroid_eq=na_access(global,SLOT_matroid_eq);
    matroid_zero_mask=na_access(local,SLOT_matroid_zero_mask);
    me_init(matroid_eq,matroid_zero_mask,params->n,params->r,beta0);
  }
}


/*
  sanity check the chirotope and read from the parser.
*/

static bool read_chirotope(na_node_p global, 
			   na_node_p local, 
			   sarray_p chirotope,
			   startopt_t startopt)
{
  int i;

  for (i=0; i <nss_count(parser.basis_signs); i++){
    int b;
    nu_sign_t sigma;
    bool ok;
    nu_signed_index_t si=nss_next(parser.basis_signs);
    
    b=nsi_index(si);
    sigma=nsi_sign(si);
    
    err_debug(DB_STARTUP,(HERE,"verifying basis %d = %s\n",b,sign_name[sigma]));
    if (startopt & SO_nonuniform){
      ok=SetAndConsequences(global,local,b,sigma);
    } else{
      ok=SetAndConsequences(global,local,b,sign2int(sigma));
    }

    if (!ok){
      err_debug(DB_INPUT,(HERE,
	       "inconsistent input, error setting basis %d to %s",b,
	       sign_name[sigma]));
      return FALSE;
    }
  }
  return TRUE;
}

static bool process_facet(na_node_p global, 
			  na_node_p local,
			  facet_p F, 
			  sarray_p boundary, 
			  oms_hook read_boundary_hook){
  int d=F->rank-1;
  int k=F->size;
  bdy_state_t sigma;
  int b;
  int *indices;
  int *elements;
  array_index_t *facet_id=na_access(global,SLOT_facet_id);
  params_p params=na_access(global,SLOT_params);
  indices=calloc(d+1,sizeof(int));
  elements=calloc(d+1,sizeof(int));
  
  sigma=fl_sign(F);

  if (sigma<0 && (k>(d+1)))
    err_warning("startup","negative facet specifications for degenerate facets may not work like you expect or at all");

  switch(F->encoding){
  case vertex_list:
    first_kset(k,d,indices);
    do {
      int j;
      int b;
      
      for (j=0; j<d; j++){
	elements[j]=F->index_set[indices[j]-1];
      }
      b=kset_cindex(F->elements,d,elements);
      sa_init(boundary,b,sigma);      
      err_debug(DB_STARTUP,(HERE,"bdy simplex %d sign %d\n",b,F->facet_id,sigma));
      read_boundary_hook(global,local,b,sigma);

    } while (next_kset(k,d,indices));

    if (k>d){
      first_kset(k,d+1,indices);

      do {
	int j;
	int b;
    
	for (j=0; j<d+1; j++){
	  elements[j]=F->index_set[indices[j]-1];
	}
	b=kset_cindex(F->elements,d+1,elements);
	err_debug(DB_STARTUP,(HERE,"boundary forcing chi[%d]=0\n",b));
	if (!SetAndConsequences(global,local,b,	SIGN_ZERO)){
	  err_debug(DB_INPUT,(HERE,"inconsistent input, error setting basis %d to 0"));
	  return(FALSE);
	}
	
      } while (next_kset(k,d+1,indices));
    }
    break;
  case basis_index:
    b=fl_bindex(F);
    sa_init(boundary,b,sigma);      
    facet_id[b]=F->facet_id;
    err_debug(DB_STARTUP,(HERE,"bdy simplex %d = facet %d sign %d\n",b,F->facet_id,sigma));
    read_boundary_hook(global,local,b,sigma);
    
    break;

  default:
    err_fatal(HERE,"unknown facet_encoding %d",F->encoding);
  }
  free(indices);
  free(elements);

  if (params->full_boundary){
    int i;
    for (i=0; i< sa_size(boundary); i++){
      if (sa_read(boundary,i)==BDY_MAYBE)
	sa_set(boundary,i,BDY_NO);
    }
  }
  return TRUE;
    
}

static bool read_boundary(na_node_p global,na_node_p local,
			  sarray_p boundary,
			  oms_hook read_boundary_hook,
			  startopt_t startopt){
  facet_p F;

  be_block_p pivots=na_access(global,SLOT_pivots);
  params_p params=na_access(global,SLOT_params);
  ridge_graph RG=na_access(local,SLOT_ridge_graph);

  for (F=fl_first(parser.facets); F; F=fl_next(F)){

    if (!process_facet(global,local,F,boundary,read_boundary_hook))
      return FALSE;

  }
  
  if (params->uniform){
    /* Now make a second pass, to initialize the ridge graph */

    for (F=fl_first(parser.facets); F;  F=fl_next(F)){

      int b,sigma;

      b=fl_bindex(F);
      sigma=fl_sign(F);

      if (sigma>0) {
	if (!rg_add_facet(RG,b,boundary,pivots)){
	  err_debug(DB_INPUT,(HERE,"failure adding facet %d initializing ridge graph",b));
	  return FALSE;
	}
      }
      
    }
  }
  return TRUE;
}

